pub type IsVar<A> = Not<IsConst<A>>;Expand description
A proof that some symbol is not a constant.
This is used to prevent collapse of propositions into constants, e.g. when doing a proof by induction.
Aliased Type
pub struct IsVar<A> { /* private fields */ }